#
# ContinuousList.py
#
# Defines an immutable list datatype using a
# continuous representation.

class ContinuousList:
    def __init__(self):
        # We use __ in the instance variable names to
        # make them private (accessible only to this class)
        self.__len = 0
        self.__items = []
        
    def access(self, index):
        assert (index >= 0 and index < self.__len)
        return self.__items[index]
    
    """
    Immutable implementation:
    """
    def append(self, value):
        res = ContinuousList ()
        res.__len = self.__len + 1
        # Need to copy first (immutable type)
        res.__items = self.__items[:] # this copies the list
        res.__items.append(value)
        return res
    
    """
    Mutable implementation:
    
    def append(self, value):
        self.__items.append(value)
        self.__len += 1
        return self
    """
        
    def length(self):
        return self.__len
    
    def __str__(self):
        res = "<"
        for index in range(self.__len):
            if not index == 0:
                res += ", "
            res += str(self.access(index))
        res += ">"
        return res