#
# Dynamic Alignment Algorithm
# Uses memoization to (more) efficiently align two sequences

GAP = '-'

# pre:  U and V are sequences (that may include gaps
# post: returns the similarity score of self and other
#          where c is the match value and g
#          is the gap penalty
# modifies: nothing
def goodnessScore (U, V, c, g):
    # You can replace this with your implementation from PS1, or
    # use the provided implementation here.
    score = 0.0
    size = min(len(U), len(V))
    
    # if sizes don't match, subtract gap penalties to fill up genome
    score -= abs(len(U) - len(V)) * g
        
    for i in range (size):
        if U[i] == GAP or V[i] == GAP:
            score -= g
        else:
            if U[i] == V[i]:
                score += c
    return score
   
# pre: U and V are strings; c and g are numbers
# post: Returns the best possible alignment of U and V given c and g
#       as two values (U', V') where U' is U with gaps inserted and V' 
#       is V with gaps inserted such that the goodnessScore of
#       U' and V' is the maximum possible goodnessScore that can
#       be obtained by adding gaps to U and V.

def bestAlignment (U, V, c, g):      
    def memoBestAlignment (U, V, c, g):
        def makeKey (U, V):
            return U + "%" + V
    
        if memo.has_key(makeKey (U,V)):
            res = memo[makeKey (U,V)]
            return res[0], res[1]
    
        if len(U) == 0 or len(V) == 0:
            while len(U) < len(V): U = U + GAP
            while len(V) < len(U): V = V + GAP
            resU = U
            resV = V
        else:
            # try with no gap
            (U0, V0) = memoBestAlignment (U[1:], V[1:], c, g)
            scoreNoGap = goodnessScore (U0, V0, c, g)
            if U[0] == V[0]: scoreNoGap += c
            
            # try inserting a gap in U (no match for V[0])
            (U1, V1) = memoBestAlignment (U, V[1:], c, g)
            scoreGapU = goodnessScore (U1, V1, c, g) - g
        
            # try inserting a gap in V (no match for U[0])
            (U2, V2) = memoBestAlignment (U[1:], V, c, g)
            scoreGapV = goodnessScore (U2, V2, c, g) - g
        
            if scoreNoGap >= scoreGapU and scoreNoGap >= scoreGapV:
                resU = U[0] + U0
                resV = V[0] + V0
            elif scoreGapU >= scoreGapV:
                resU = GAP + U1
                resV = V[0] + V1
            else:
                resU = U[0] + U2
                resV = GAP + V2
        
        memo[makeKey(U,V)] = [resU, resV]
        return resU, resV
    
    memo = {}
    return memoBestAlignment (U, V, c, g)
        
# pre: U and V are sequences
# post: returns the best goodness score obtainable between U and V
#     with match score c and gap penalty g
def bestScore (U, V, c, g):
    (Ub, Vb) = bestAlignment (U, V, c, g)
    return goodnessScore (Ub, Vb, c, g)
